# -*- coding: utf-8 -*-
"""
Created on Tue Aug 28 21:12:23 2018

@author: limingfan
"""

import os
import time

import logging


class ModelSettingsTemplate(object):
    """
    """
    def __init__(self):

        # model graph
        self.model_tag = None     # str
        self.model_graph = None   # ModelGraph class with static methods
        self.is_train = None      # bool
        
        # model graph macro parameters
        #
        
        # other macro parameters
        #

        # train
        self.gpu_available = "0"      # could be specified in args
        self.gpu_batch_split = None   # list, for example, [12, 20]; if None, batch split evenly
        self.gpu_mem_growth = True
        self.log_device = False
        self.soft_placement = True

        self.num_epochs = 100     
        self.batch_size = 10
        self.batch_size_eval = 10
        
        self.reg_lambda = 0.000001  # 0.0, 0.0001
        self.grad_clip = 8.0  # 0.0, 5.0, 8.0, 2.0
        self.keep_prob = 0.8  # 1.0, 0.7, 0.5
        
        self.optimizer_type = 'adam'  # adam, momentum, sgd
        self.momentum = 0.9
        self.learning_rate_base = 0.001   #
        self.ratio_decay = 0.9
        self.patience_decay = 3000
        self.learning_rate_minimum = 0.00001
        
        self.save_period_batch = 100
        self.valid_period_batch = 100
        #
        
        """
        # inputs/outputs
        self.vs_str_multi_gpu = "vs_multi_gpu"
        #
        self.inputs_predict_name = ['src_seq:0', 'src_seq_mask:0']
        self.outputs_predict_name = ['logits:0']
        self.pb_outputs_name = ['logits']
                
        self.inputs_train_name = ['src_seq:0', 'src_seq_mask:0',
                                  'dcd_seq:0', 'dcd_seq_mask:0',
                                  'dcd_seq:0', 'dcd_seq_mask:0',
                                  'labels_seq:0', 'labels_mask:0']
        self.outputs_train_name = ['logits:0']
        self.use_metric = True
        
        self.debug_tensors_name = ['loss/loss_batch:0',
                                   'logits:0',
                                   'logits_normed:0',
                                   'preds:0'
                                   #'encoder/encoder_rnn_1/bw/bw/sequence_length:0',
                                   #'inputs_len:0'
                                   ]
        """
        
        #
        # save and log, if not set, default values will be used.
        self.base_dir = '.'
        self.model_dir = None
        self.model_name = None
        self.pb_file = None
        self.log_dir = None
        self.log_path = None
        #
    
    def check_settings(self):
        """ assert and make directories
        """        
        # assert         
        assert self.is_train is not None, 'is_train not assigned'               
        assert self.model_tag is not None, 'model_tag is None'
        
        # gpu
        self.num_gpu = len(self.gpu_available.split(","))
        if self.num_gpu > 1:
            #
            if self.gpu_batch_split is None:
                self.gpu_batch_split = [self.batch_size//self.num_gpu] * self.num_gpu
            #
            str_info = "make sure that self.num_gpu == len(self.gpu_batch_split)"
            assert self.num_gpu == len(self.gpu_batch_split), str_info
            str_info = "make sure that self.batch_size == sum(self.gpu_batch_split)"
            assert self.batch_size == sum(self.gpu_batch_split), str_info
            
        # directories
        if self.model_dir is None:
            self.model_dir = os.path.join(self.base_dir, 'model_' + self.model_tag)
        if self.log_dir is None: self.log_dir = os.path.join(self.base_dir, 'log')
        #
        if not os.path.exists(self.base_dir): os.mkdir(self.base_dir)
        if not os.path.exists(self.model_dir): os.mkdir(self.model_dir)
        if not os.path.exists(self.model_dir + '_best'): os.mkdir(self.model_dir + '_best')
        if not os.path.exists(self.log_dir): os.mkdir(self.log_dir)
        #
        # files
        if self.model_name is None: self.model_name = 'model_' + self.model_tag
        if self.pb_file is None: self.pb_file = os.path.join(self.model_dir + '_best',
                                                             self.model_name + '.pb')
        #
        # logger
        str_datetime = time.strftime("%Y-%m-%d-%H-%M")       
        if self.log_path is None: self.log_path = os.path.join(
                self.log_dir, self.model_name + "_" + str_datetime +".txt")
        #
        self.logger = logging.getLogger(self.log_path)  # use log_path as log_name
        self.logger.setLevel(logging.INFO)
        handler = logging.FileHandler(self.log_path)
        handler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        # formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        # self.logger.info('test')
        #
        self.display()
        #
        
    def create_or_reset_log_file(self):        
        with open(self.log_path, 'w', encoding='utf-8'):
            pass
        
    def close_logger(self):
        for item in self.logger.handlers: item.close()
        
    def display(self):
        
        print()
        for name,value in vars(self).items():
            if not isinstance(value, (int, float, str, bool, list, dict, tuple)):
                continue
            print(str(name) + ': ' + str(value))
        print()
        
    def trans_info_to_dict(self):
                
        info_dict = {}
        for name,value in vars(self).items():
            if not isinstance(value, (int, float, str, bool, list, dict, tuple)):
                continue
            info_dict[str(name)] = value        
        return info_dict
    
    def assign_info_from_dict(self, info_dict):
        
        pass
    
        
if __name__ == '__main__':
    
    sett = ModelSettingsTemplate()
    
    sett.model_tag = 'cnn'
    sett.is_train = False
    
    #print(dir(sett))
    #l = [i for i in dir(sett) if inspect.isbuiltin(getattr(sett, i))]
    #l = [i for i in dir(sett) if inspect.isfunction(getattr(sett, i))]
    #l = [i for i in dir(sett) if not callable(getattr(sett, i))]
    
    sett.check_settings()
    
    print(sett.__dict__.keys())
    print()
    
    sett.close_logger()
    


# -*- coding: utf-8 -*-
"""
Created on Sat Apr 13 14:53:37 2019

@author: limingfan
"""

import os
import random
import numpy as np

import threading

from Zeras.vocab import Vocab

#
class threadsafe_iter:
    """Takes an iterator/generator and makes it thread-safe by
    serializing call to the `next` method of given iterator/generator.
    """

    def __init__(self, it):
        self.it = it
        self.lock = threading.Lock()

    def __iter__(self):
        return self

    def __next__(self): # python3
        with self.lock:
            return self.it.__next__()
  
     # def next(self): # python2
     #     with self.lock:
     #       return self.it.next()
     
def threadsafe_generator(f):
    """A decorator that takes a generator function and makes it thread-safe.
    """
    def g(*a, **kw):
        return threadsafe_iter(f(*a, **kw))
    return g

#
def get_examples_generator(data_files = []):
    """
    """
    def examples_generator(single_pass=True):
        """
        """
        vocab = Vocab()
        for idx in range(10):
            vocab.add(str(idx))
        #
        count_examples = 0
        while True:
            src_seq_token = [str(random.randint(0, 9)) for _ in range(random.randint(1, 10))]
            src_seq = vocab.convert_tokens_to_ids(src_seq_token)
            
            if single_pass and count_examples == 600: break
            count_examples += 1
        
            yield src_seq
            #
    #
    return examples_generator
    #

def batch_std_transor(list_examples, max_seq_len = 20, tgt_seq_len = 12):
    """
    """    
    start_id = 2
    end_id = 3
    #
    num_examples = len(list_examples)
    list_len = [len(item) for item in list_examples]
    max_len = min(max_seq_len, max(list_len))
    
    src_seq = np.zeros((num_examples, max_len), dtype=np.int32)
    src_seq_mask = np.zeros((num_examples, max_len), dtype=np.int32)
    
    dcd_seq = np.zeros((num_examples, tgt_seq_len), dtype=np.int32)
    dcd_seq_mask = np.zeros((num_examples, tgt_seq_len), dtype=np.int32)
    lbl_seq = np.zeros((num_examples, tgt_seq_len), dtype=np.int32)
    lbl_seq_mask = np.zeros((num_examples, tgt_seq_len), dtype=np.int32)
    
    for eid in range(num_examples):
        
        dcd_seq[eid, 0] = start_id
        dcd_seq_mask[eid, 0] = 1
        
        src_len = list_len[eid]        
        for tid in range(src_len):
            src_seq[eid, tid] = list_examples[eid][tid]
            src_seq_mask[eid, tid] = 1
            
            if tid >= tgt_seq_len: continue
            lbl_seq[eid, tid] = list_examples[eid][tid]
            lbl_seq_mask[eid, tid] = 1
            
            if tid+1 >= tgt_seq_len: continue
            dcd_seq[eid, tid+1] = list_examples[eid][tid]
            dcd_seq_mask[eid, tid+1] = 1

        #
        if src_len >= tgt_seq_len: continue
        lbl_seq[eid, src_len] = end_id
        lbl_seq_mask[eid, src_len] = 1
        
        if src_len+1 >= tgt_seq_len: continue
        dcd_seq[eid, src_len+1] = end_id
        dcd_seq_mask[eid, src_len] = 1
        #
        
    #
    return src_seq, src_seq_mask, dcd_seq, dcd_seq_mask, lbl_seq, lbl_seq_mask
    #
    
#
if __name__ == "__main__":
    
    #
    vocab = Vocab()
    for idx in range(10):
        vocab.add(str(idx))
    
    dir_vocab = "./vocab"
    if not os.path.exists(dir_vocab): os.mkdir(dir_vocab)
    token_file = os.path.join(dir_vocab, "vocab_tokens.txt")
    vocab.save_tokens_to_file(token_file)
    
    #
    data_gen = get_examples_generator()
    it = data_gen()
    print(next(it))
    print(next(it))
    
    a = [next(it) for _ in range(3)]
    print(a)
    
    batch = batch_std_transor(a)
    print(batch)
    


# -*- coding: utf-8 -*-
"""
Created on Sat Apr  6 21:24:16 2019

@author: limingfan
"""

from Zeras.layers import Dense
from Zeras.layers import build_module_copies

import tensorflow as tf


tf.reset_default_graph()


dc = build_module_copies(Dense, (10, 10), 3)
print(dc)
print(dc[0].wb)


#
a = [0,1,2,3,4]

data_iter = a.__iter__()

print(data_iter)

while True:
    print(next(data_iter))




# -*- coding: utf-8 -*-
"""
Created on Sat Sep  1 17:14:19 2018

@author: limingfan
"""

import numpy as np
import tensorflow as tf


class Dense():
    """
    """
    def __init__(self, input_size, output_size, weight_mat=None,
                 use_bias=True, bias_init_value=0.0, scope="dense"):
        """
        """
        wb = create_dense_vars(input_size, output_size,
                               weight_mat=weight_mat, use_bias=use_bias,
                               bias_init_value=bias_init_value, scope=scope)
        #
        self.wb = wb
    
    def __call__(self, inputs, transpose_b=False):
        """
        """
        out = dense_with_vars(inputs, self.wb, transpose_b=transpose_b)
        return out

class LayerNorm():
    """
    """
    def __init__(self, num_units, epsilon=1e-6, scope="layer_norm"):
        """
        """
        with tf.variable_scope(scope):
            self.beta = tf.get_variable('layer_norm_beta', [num_units],
                                        initializer=tf.ones_initializer(),
                                        trainable=True)
            self.gamma = tf.get_variable('layer_norm_gamma', [num_units],
                                         initializer=tf.zeros_initializer(),
                                         trainable=True)
            self.eps = epsilon
    
    def __call__(self, x):
        """
        """
        mean, std = tf.nn.moments(x, [-1], keep_dims=True, name='moments')        
        return self.beta * (x - mean)/ (std + self.eps) + self.gamma
    
class Dropout():
    """
    """
    def __init__(self, keep_prob):
        self.keep_prob = keep_prob
        
    def __call__(self, x):        
        return tf.nn.dropout(x, self.keep_prob)
    
#
def dense_with_w(inputs, hidden, weights, transpose_b=False):
    """
    """
    shape_list = inputs.get_shape().as_list()
    if len(shape_list) == 2:
        out = tf.matmul(inputs, weights, transpose_b=transpose_b)
        return out
    #
    shape = tf.shape(inputs)
    out_shape = [shape[idx] for idx in range(len(shape_list) - 1)] + [hidden]
    input_size = shape_list[-1]
    flat_inputs = tf.reshape(inputs, [-1, input_size])
    out = tf.matmul(flat_inputs, weights, transpose_b = transpose_b)
    out = tf.reshape(out, out_shape)
    return out

def create_dense_vars(input_size, output_size, weight_mat=None,
                      use_bias=True, bias_init_value=0.0, scope="dense"):
    """
    """
    with tf.variable_scope(scope):
        if weight_mat is None:
            W = tf.get_variable("kernel", [input_size, output_size],
                                initializer = tf.variance_scaling_initializer())
        else:
            W = weight_mat
        if use_bias:
            b = tf.get_variable("bias", [output_size],
                                initializer = tf.constant_initializer(bias_init_value))
        else:
            b = None
    #
    return W, b

def dense_with_vars(inputs, Wb, transpose_b=False):
    """
    """
    shape_list = inputs.get_shape().as_list()
    if len(shape_list) == 2:
        out = tf.matmul(inputs, Wb[0], transpose_b=transpose_b)
        if Wb[1] is not None: out = tf.nn.bias_add(out, Wb[1])
        return out
    #
    input_size = shape_list[-1]
    shape = tf.shape(inputs)
    if transpose_b:
        output_size = Wb[0].get_shape().as_list()[0]
    else:
        output_size = Wb[0].get_shape().as_list()[1]
    out_shape = [shape[idx] for idx in range(len(shape_list) - 1)] + [output_size]
    flat_inputs = tf.reshape(inputs, [-1, input_size])
    out = tf.matmul(flat_inputs, Wb[0], transpose_b=transpose_b)
    if Wb[1] is not None: out = tf.nn.bias_add(out, Wb[1])
    out = tf.reshape(out, out_shape)
    return out

def layer_norm(x, scope=None):
    num_units = tf.shape(x)[-1]
    ln = LayerNorm(num_units, scope=scope)
    return ln(x)
    
# 
def build_module_copies(module_class, class_args, N, scope="module_copies"):
    """ module_class: must be some class
        class_args: be the args for module_class init function
    """
    list_copies = []
    with tf.variable_scope(scope):
        for idx in range(N):
            with tf.variable_scope("copy_%d" % idx):
                module_copied = module_class(*class_args)
                list_copies.append(module_copied)
    return list_copies

#
def qkv_att_layer(query, key, value, mask_mat=None, keep_prob=1.0):
    """ batch_major
        query: [B, TQ, DQ]
        key: [B, TK, DK]    # DQ = DK
        value: [B, TV, DV]  # TK = TV
        mask_mat: [B, TQ, TK], or [1, TQ, TK]
        
        return: [B, TQ, DV]
        3-dim, or higher-dim
    """
    dim = query.get_shape().as_list()[-1]
    att_mat = tf.matmul(query, key, transpose_b=True) / (dim ** 0.5)
    #
    if mask_mat is not None:
        att_mat = tf.add(att_mat, 1e16 * (mask_mat - 1) )  # -inf   # [B, TQ, TM]
    #
    logits = tf.nn.softmax(att_mat)
    logits = tf.nn.dropout(logits, keep_prob)
    outputs = tf.matmul(logits, value)   # [B, TQ, DV]
    return outputs, logits

#
class MultiHeadAttention():
    """
    """
    def __init__(self, num_heads, num_units, keep_prob=1.0, scope="multi_head"):
        """
        """
        self.keep_prob = keep_prob
        self.num_heads = num_heads
        self.num_units = num_units
        self.dim_all = self.num_heads * self.num_units
        
        self.attention = None
        
        d_model = num_heads * num_units
        with tf.variable_scope(scope):
            self.dense_query = Dense(d_model, d_model, scope="dense_query")
            self.dense_key = Dense(d_model, d_model, scope="dense_key")
            self.dense_value = Dense(d_model, d_model, scope="dense_value")
            self.dense_trans = Dense(d_model, d_model, scope="dense_trans")
            
        
    def __call__(self, query, key, value, mask_mat=None):
        """
        """
        qd = self.dense_query(query)
        kd = self.dense_key(key)
        vd = self.dense_value(value)
        #
        spq = tf.shape(query)
        batch_size = spq[0]
        q_len = spq[1]
        k_len = tf.shape(key)[1]
        # v_len = spq[1]
        #
        qs = tf.reshape(qd, [batch_size, q_len, self.num_heads, self.num_units])
        ks = tf.reshape(kd, [batch_size, k_len, self.num_heads, self.num_units])
        vs = tf.reshape(vd, [batch_size, k_len, self.num_heads, self.num_units])
        #
        qe = tf.transpose(qs, [0, 2, 1, 3])   # to [B, H, T, D]
        ke = tf.transpose(ks, [0, 2, 1, 3])
        ve = tf.transpose(vs, [0, 2, 1, 3])
    
        # qkv
        if mask_mat is None:
            mask_mat_e = None
        else:
            mask_mat_e = tf.expand_dims(mask_mat, 1)
        #
        out, att = qkv_att_layer(qe, ke, ve, mask_mat_e, keep_prob=1.0)
        #
        self.attention = att
        #
        
        # concat
        out_c = tf.transpose(out, [0, 2, 1, 3])           # to [B, T, H, D]
        out_c = tf.reshape(out, [batch_size, q_len, self.dim_all])    
        
        # linear
        out_d = self.dense_trans(out_c)
        return out_d
    
class PositionwiseFeedForward():
    """
    """
    def __init__(self, num_dim_all, dim_middle, keep_prob, scope="pwff"):
        """
        """        
        with tf.variable_scope(scope):
            self.d1 = Dense(num_dim_all, dim_middle, scope="ff_d1")
            self.d2 = Dense(dim_middle, num_dim_all, scope="ff_d2")
            self.dropout = Dropout(keep_prob)
        
    def __call__(self, x):
        x = tf.nn.relu(self.d1(x))
        x = self.d2(self.dropout(x))
        return x

#
class SublayerWrapper():
    """
    """
    def __init__(self, num_units, keep_prob, sublayer_class, class_args,
                 scope="sublayer_wrapper"):
        """
        """
        with tf.variable_scope(scope):
            self.layer_norm = LayerNorm(num_units, scope="sublayer_wrapper")
            self.dropout = Dropout(keep_prob)
            self.sublayer = sublayer_class(*class_args)
    
    def __call__(self, x, sublayer_invoker):
        """
        """
        return x + self.dropout(sublayer_invoker(self.layer_norm(x)))
    
#
def get_mask_mat_from_mask_seq(mask_a, mask_b):
    """ mask_a: [B, TA]
    """
    mask_ae = tf.cast(tf.expand_dims(mask_a, 2), tf.float32)  # [B, TA, 1]
    mask_be = tf.cast(tf.expand_dims(mask_b, 1), tf.float32)  # [B, 1, TB]
    mask = mask_ae * mask_be    # [B, TA, TB]
    return mask

#
def get_mask_mat_subsequent(size, name="mask_subsequent"):
    """ subsequent mask
    """
    
    """
    mask_mat = np.zeros((1, size, size), dtype = np.float32)
    for idx in range(size):
        for idy in range(size):
            if idx <= idy: mask_mat[0, idx, idy] = 1.0
    #
    mask_tensor = tf.get_variable(name, shape = (1, size, size),
                                  initializer = tf.constant_initializer(mask_mat),
                                  trainable = False)
    """
    #
    mask_tensor = tf.constant(1.0, shape = (1, size, size), dtype=tf.float32)
    mask_tensor = tf.linalg.band_part(mask_tensor,
                                      num_lower = -1,
                                      num_upper = 0,
                                      name = name)
    return mask_tensor

def get_list_subs_masks(max_len, name="subs_mask"):
    """ subsequent masks
    """    
    list_masks = []
    for step in range(max_len):
        subs_mask = get_mask_mat_subsequent(step+1, name = name+"_%d" % step)
        list_masks.append(subs_mask)
    return list_masks

def get_list_dcd_crs_masks(src_mask_seq, max_len):
    """ decoder cross masks
    """    
    list_masks = []
    mask_be = tf.cast(tf.expand_dims(src_mask_seq, 1), tf.float32)  # [B, 1, TB]
    for step in range(max_len):        
        crs_mask = tf.tile(mask_be, [1, step+1, 1])
        list_masks.append(crs_mask)
    return list_masks
    
#
def get_position_emb_mat(max_seq_len, posi_emb_dim, posi_emb_model,
                         name="position_embeddings"):
    """
    """
    d_model_recip_2 = 2.0 / posi_emb_model
    
    arg_mat = np.zeros((max_seq_len, posi_emb_dim), dtype=np.float32)
    for idx in range(max_seq_len):
        for idm in range(posi_emb_dim):
            arg_mat[idx, idm] = idx * 1e-4**(d_model_recip_2 * idm)
    #
    pe_sin = np.sin(arg_mat)
    pe_cos = np.cos(arg_mat)
    #    
    pe_sin = np.expand_dims(pe_sin, -1)  
    pe_cos = np.expand_dims(pe_cos, -1)
    pe_all = np.concatenate([pe_sin, pe_cos], -1)  # (T, D, 2)
    #
    pe_all = np.reshape(pe_all, [max_seq_len, -1])
    pe_all = pe_all[:, 0:posi_emb_dim]
    #
    # tf.Tensor
    pe_mat = tf.get_variable(name, shape = (max_seq_len, posi_emb_dim),
                             initializer = tf.constant_initializer(pe_all),
                             trainable = False)
        
    return pe_mat

def get_emb_positioned(x, token_emb, position_emb):
    """ x: [None, None]
    """
    posi = tf.range(tf.shape(x)[-1])
    
    seq_emb_t = tf.nn.embedding_lookup(token_emb, x)
    seq_emb_p = tf.nn.embedding_lookup(position_emb, posi)
    
    return seq_emb_t + seq_emb_p

#
def gelu(x):
    cdf = 0.5 * (1.0 + tf.tanh((0.79788456 * (x + 0.044715 * tf.pow(x, 3)) )))
    return x * cdf




# -*- coding: utf-8 -*-

import os

from Zeras.vocab import Vocab

from task_copy_model_settings import ModelSettings
import task_copy_model_utils as model_utils

import argparse


def parse_args():
    """
    Parses command line arguments.
    """
    parser = argparse.ArgumentParser('task_copy')
    parser.add_argument('--mode', choices=['train', 'eval', 'predict', 'debug'],
                        default = 'train', help = 'run mode')
    #
    parser.add_argument('--note', type=str, default = 'note_something',
                        help = 'make some useful notes')
    parser.add_argument('--debug', type=int, default = 0,
                        help = 'debug or not (using debug data or not)')
    parser.add_argument('--gpu', type=str, default = '0',
                        help = 'specify gpu device')
    #
    parser.add_argument('--max_batches_eval', type=int, default = 10,
                        help = 'specify how many batches go through eval')
    #
    model_related = parser.add_argument_group('model related settings')
    model_related.add_argument('--model_tag', type=str,
                               default = 'transformer', help='model_tag')
    model_related.add_argument('--base_dir', type=str,
                               default = 'task_copy_result',
                               help='base directory for saving models')
    #
    vocab_related = parser.add_argument_group('vocab related settings')
    vocab_related.add_argument('--emb_file', type=str, default = None,
                               help='pretrained embeddings file')
    vocab_related.add_argument('--tokens_file', type=str,
                               default = './vocab/vocab_tokens.txt',
                               help='tokens file')
    
    return parser.parse_args()

#  
if __name__ == '__main__':
    
    args = parse_args()
    run_mode = args.mode
    #
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu
    #
    model_tag = args.model_tag
    #
    if model_tag.startswith('transformer'):
        from model_graph_transformer import ModelGraph
    
    #
    # settings & vocab
    settings = ModelSettings()
    settings.model_tag = model_tag
    settings.model_graph = ModelGraph
    settings.gpu_available = args.gpu
    #
    vocab = Vocab()    
    vocab.add_tokens_from_file(args.tokens_file)
    vocab.load_pretrained_embeddings(args.emb_file)
    vocab.emb_dim = settings.emb_dim
    settings.vocab = vocab
    #
    if run_mode == 'predict':
        settings.is_train = False
    else:
        settings.is_train = True
    #
    settings.base_dir = args.base_dir
    settings.check_settings()
    settings.create_or_reset_log_file()
    settings.logger.info('running with args : {}'.format(args))
    settings.logger.info(settings.trans_info_to_dict())
    
    #
    # run
    if run_mode == 'debug':
        model_utils.do_debug(settings, args)
    elif run_mode == 'train':
        model_utils.do_train_and_valid(settings, args)
    elif run_mode == 'eval':
        model_utils.do_eval(settings, args)
    elif run_mode == 'predict':
        model_utils.do_predict(settings, args)
    else:
        print('NOT supported mode. supported modes: debug, train, eval, and predict.')
    #
    
    


# -*- coding: utf-8 -*-

from Zeras.layers import Dense, LayerNorm
from Zeras.layers import MultiHeadAttention, PositionwiseFeedForward
from Zeras.layers import SublayerWrapper
from Zeras.layers import build_module_copies

import decoding_method as dm


class EncoderDecoder():
    """
    """
    def __init__(self, encoder, decoder,
                 src_emb_trans, tgt_emb_trans, generator):
        
        self.encoder = encoder
        self.decoder = decoder
        self.src_emb_trans = src_emb_trans
        self.tgt_emb_trans = tgt_emb_trans
        self.generator = generator
        
    def forward(self, src, src_mask, dcd_inputs, dcd_mask, crs_mask): 
        """ source, decode, cross
        """
        return self.decode(dcd_inputs, dcd_mask, self.encode(src, src_mask), crs_mask)
        
    def encode(self, src, src_mask):
        return self.encoder(self.src_emb_trans(src), src_mask)
    
    def decode(self, dcd_inputs, dcd_mask, memory, crs_mask):
        """ decode one-step
        """
        return self.decoder(self.tgt_emb_trans(dcd_inputs), dcd_mask, memory, crs_mask)
    
    def do_greedy_decoding(self, src, src_mask, max_len,
                           subs_masks, dcd_crs_masks, start_symbol_id):
        """ decode max_len steps
        """
        return dm.do_greedy_decoding(self, src, src_mask, max_len, 
                                     subs_masks, dcd_crs_masks, start_symbol_id)
    
    def do_beam_search_decoding(self, src, src_mask, max_len,
                                subs_masks, dcd_crs_masks, start_symbol_id):
        """ decode max_len steps
        """
        return dm.do_beam_search_decoding(self, src, src_mask, max_len,
                                          subs_masks, dcd_crs_masks, start_symbol_id)
    
#   
class Generator():
    """
    """
    def __init__(self, d_model, vocab_size, emb_mat=None, scope="proj"):
        """
        """        
        self.emb_mat = emb_mat
        self.proj = Dense(d_model, vocab_size, weight_mat=emb_mat, scope=scope)
    
    def forward(self, x):
        if self.emb_mat is None:
            out = self.proj(x, transpose_b=False)
        else:
            out = self.proj(x, transpose_b=True)
        return out
        
#  
class Encoder():
    """
    """
    def __init__(self, num_layers, layer_module, module_args, scope="encoder"):
        
        self.layers = build_module_copies(layer_module, module_args, num_layers,
                                          scope = scope)
        self.layer_norm = LayerNorm(self.layers[-1].num_all, scope=scope)
    
    def __call__(self, x, mask):
        """ x: seq_embedded
        """        
        for layer in self.layers:
            x = layer(x, mask)
        return self.layer_norm(x)

class Decoder():
    """
    """
    def __init__(self, num_layers, layer_module, module_args, scope="decoder"):
        
        self.layers = build_module_copies(layer_module, module_args, num_layers,
                                          scope = scope)
        self.layer_norm = LayerNorm(self.layers[-1].num_all, scope=scope)
    
    def __call__(self, x, dcd_mask, memory, crs_mask):
        """ x: seq_embedded
        """        
        for layer in self.layers:
            x = layer(x, dcd_mask, memory, crs_mask)
        return self.layer_norm(x)
    
#
class EncoderLayer():
    """
    """
    def __init__(self, num_dim_all, att_args, ffd_args, keep_prob):
        
        self.num_all = num_dim_all
        self.sublayer_0 = SublayerWrapper(num_dim_all, keep_prob,
                                          MultiHeadAttention, att_args,
                                          scope = "sublayer_0")
        self.sublayer_1 = SublayerWrapper(num_dim_all, keep_prob,
                                          PositionwiseFeedForward, ffd_args,
                                          scope = "sublayer_1")
        
    def __call__(self, x, mask):
        
        x = self.sublayer_0(x, lambda x: self.sublayer_0.sublayer(x, x, x, mask))
        x = self.sublayer_1(x, self.sublayer_1.sublayer)
        return x
    
class DecoderLayer():
    """
    """
    def __init__(self, num_dim_all, att_args, src_args, ffd_args, keep_prob):
        
        self.num_all = num_dim_all
        self.sublayer_0 = SublayerWrapper(num_dim_all, keep_prob,
                                          MultiHeadAttention, att_args,
                                          scope = "sublayer_0")
        self.sublayer_1 = SublayerWrapper(num_dim_all, keep_prob,
                                          MultiHeadAttention, src_args,
                                          scope = "sublayer_1")
        self.sublayer_2 = SublayerWrapper(num_dim_all, keep_prob,
                                          PositionwiseFeedForward, ffd_args,
                                          scope = "sublayer_2")
        
    def __call__(self, x, dcd_mask, memory, crs_mask):
        
        m = memory
        x = self.sublayer_0(x, lambda x: self.sublayer_0.sublayer(x, x, x, dcd_mask))
        x = self.sublayer_1(x, lambda x: self.sublayer_1.sublayer(x, m, m, crs_mask))
        x = self.sublayer_2(x, self.sublayer_2.sublayer)
        return x
        
#

    


# -*- coding: utf-8 -*-
"""
Created on Thu Jan 31 05:08:20 2019

@author: limingfan
"""

import numpy as np

from Zeras.data_batcher import DataBatcher
from Zeras.model_wrapper import ModelWrapper

import task_copy_data_set as data_set


#
def eval_process(model, eval_batcher, args, flag_score):
    #
    max_batches_eval = args.max_batches_eval
    #
    loss_aver, metric_aver = 0.0, 0.0
    count = 0
    while True:
        batch = eval_batcher.get_next_batch()  
        #
        if batch is None: break
        if count == max_batches_eval: continue  #
        #
        count += 1
        # print(count)
        #
        results, loss, metric = model.run_eval_one_batch(batch)
        loss_aver += loss
        metric_aver += metric
        # print(loss)
        # print(metric)
        # print(results)
        # 
    #
    loss_aver /= count
    # metric_aver /= count
    metric_aver = 100 - loss_aver
    #
    model.logger.info('eval finished, with total num_batches: %d' % count)
    # model.logger.info('loss_aver, metric_aver: %g, %g' % (loss_aver, metric_aver))
    #
    eval_score = {}
    return eval_score, loss_aver, metric_aver
    #
    
def do_eval(settings, args):
    #
    # model
    model = ModelWrapper(settings)
    model.prepare_for_train_and_valid()
    model.assign_dropout_keep_prob(1.0)
    #
    # data
    data_batcher = DataBatcher(data_set.get_examples_generator(), data_set.batch_std_transor,
                               settings.batch_size, single_pass = True, worker_type="thread")
    #
    # eval
    eval_score, loss_aver, metric_aver = eval_process(model, data_batcher, args, 1)
    #
    print('loss_aver, metric_aver: %g, %g' % (loss_aver, metric_aver))
    model.logger.info('loss_aver, metric_aver: %g, %g' % (loss_aver, metric_aver))
    model.logger.info('{}'.format(eval_score))
    #
    
def do_train_and_valid(settings, args):
    #
    # model
    model = ModelWrapper(settings)
    model.prepare_for_train_and_valid()
    #    
    # data
    data_batcher = DataBatcher(data_set.get_examples_generator(), data_set.batch_std_transor,
                               settings.batch_size, single_pass = False, worker_type="thread")
    eval_period = settings.valid_period_batch
    #
    loss = 10000.0
    best_metric_val = 0
    last_improved = 0
    lr = model.learning_rate_base
    #
    count = 0
    while True:
        #
        # eval
        if count % eval_period == 0:
            model.logger.info("training curr batch, loss, lr: %d, %g, %g" % (count, loss, lr) )
            #
            model.save_ckpt(model.model_dir, model.model_name, count)
            model.assign_dropout_keep_prob(1.0)
            #
            model.logger.info('evaluating after num_batches: %d' % count)
            eval_batcher = DataBatcher(data_set.get_examples_generator(), data_set.batch_std_transor,
                                       settings.batch_size, single_pass = True, worker_type="thread")
            #
            eval_score, loss_aver, metric_val = eval_process(model, eval_batcher, args, 0)
            model.logger.info("eval loss_aver, metric, metric_best: %g, %g, %g" % (
                    loss_aver, metric_val, best_metric_val) )
            #
            # save best
            if metric_val >= best_metric_val:  # >=
                best_metric_val = metric_val
                last_improved = count
                # ckpt
                model.logger.info('a new best model, saving ...')
                model.save_ckpt_best(model.model_dir + '_best', model.model_name, count)
                # pb
                model.save_graph_pb_file(model.pb_file)
                #
            
            # decay
            if count - last_improved >= model.patience_decay:
                lr *= model.ratio_decay
                model.assign_learning_rate(lr)
                last_improved = count
                model.logger.info('learning_rate decayed after num_batches: %d' % count)
                model.logger.info('current learning_rate %g' % lr)
                #
                if lr < model.learning_rate_minimum:
                    model.logger.info('current learning_rate < learning_rate_minimum, stop training')
                    break
                #
            #
            model.assign_dropout_keep_prob(settings.keep_prob)
            model.logger.info('')
            #
        #
        # train
        batch = data_batcher.get_next_batch()  
        # if batch is None: break
        count += 1
        # print(count)        
        #
        loss = model.run_train_one_batch(batch)   # just for train
        # print(loss)
        # model.logger.info("training curr batch, loss, lr: %d, %g, %g" % (count, loss, lr) )
        #
        # debugs = model.run_debug_one_batch(batch.data)
        # print(debugs[0])
        # print(debugs[1][0])
        # print(debugs[2][0])
        # print(debugs[3][0])
        # print(batch.data[8][0])
        #
    #
    model.logger.info("training finshed with total num_batches: %d" % count)
    #
    
def do_debug(settings, args):
    #
    # model
    model = ModelWrapper(settings)
    model.prepare_for_train_and_valid()
    #    
    # data
    data_batcher = DataBatcher(data_set.get_examples_generator(), data_set.batch_std_transor,
                               settings.batch_size, single_pass = False, worker_type="thread")
    #
    count = 0
    while True:        
        #
        batch = data_batcher.get_next_batch()  
        # if batch is None: break
        if count == 5: break
        count += 1
        # print(count)
        #
        loss = model.run_train_one_batch(batch)
        print(loss)
        #
        debugs = model.run_debug_one_batch(batch)
        print(debugs[0])
        print(debugs[1][0])
        print(debugs[2][0])
        #
    #
    print('finished, with total num_batches:')
    print(count)
    #
    
def do_predict(settings, args):
    #
    # model
    model = ModelWrapper(settings)
    model.prepare_for_prediction()
    # model.prepare_for_train_and_valid()
    # model.assign_dropout_keep_prob(1.0)
    
    #
    # data
    data_batcher = DataBatcher(data_set.get_examples_generator(), data_set.batch_std_transor,
                               settings.batch_size, single_pass = True, worker_type="thread")
    #
    count = 0
    while True:
        batch = data_batcher.get_next_batch()  
        #
        if batch is None: break
        if count == args.max_batches_eval: continue  #
        #
        count += 1
        print(count)
        #
        #debugs = model.run_debug_one_batch(batch.data)
        #print(debugs[0])
        #print(debugs[1])
        print("batch data:")
        print(batch[0])
        print("batch data end")
        #
        results = model.predict_from_batch(batch)[0]
        # results = model.run_predict_one_batch(batch.data)
        print("results:")
        print(np.argmax(results, -1) )
        print("results end")
        print()
        #
        # print(batch.data[8][0])
        #
    #
    model.logger.info('prediction finished, with total num_batches: %d' % count)
    #
    